-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("Aloft")
local SML = AceLibrary("SharedMedia-1.0")
local roster = AceLibrary:HasInstance("Roster-2.1") and AceLibrary("Roster-2.1") or nil

-----------------------------------------------------------------------------

Aloft = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
Aloft.version = "1.0"
Aloft.hasIcon = "Interface\\Icons\\Spell_Frost_Wisp"	-- Placeholder
Aloft.defaultPosition = 'LEFT'
Aloft.hideWithoutStandby = true
Aloft.independentProfile = true

-----------------------------------------------------------------------------

local function FormatLargeValue(value)
	if value < 10000 then
		-- Force it to an integer
		return string.format("%d", value)
	elseif value < 100000 then
		return string.format("%.1fk", value / 1000)
	elseif value < 1000000 then
		return string.format("%dk", value / 1000)
	else
		return string.format("%.2fm", value / 1000000)
	end
end

-----------------------------------------------------------------------------


local numberOfWorldChildren = 0
local nameplateList = { }
local outlineLookup = { ["None"] = nil, ["Normal"] = "OUTLINE", ["Thick"] = "THICKOUTLINE" }

local castBarTextFrame = nil
local castBarTimeTextFrame = nil

local healthTextMethods =
{
	["Hidden"] = function(frame, healthBar) return "" end,
	["Percent"] = function(frame, healthBar) return string.format("%d%%", floor(100 * frame.aloftData.healthBarValue / select(2, healthBar:GetMinMaxValues()))) end,
	["Group Health"] =	function(frame, healthBar) 
							local maxValue = select(2, healthBar:GetMinMaxValues())
							if maxValue == 100 then 
								return "" 
							end
							return FormatLargeValue(frame.aloftData.healthBarValue)
						end,
	["Group Health Deficit"] =	function(frame, healthBar) 
									local maxValue = select(2, healthBar:GetMinMaxValues())
									if maxValue == 100 then 
										return "" 
									end
									local value = frame.aloftData.healthBarValue
									if value == maxValue then 
										return "" 
									end
									return FormatLargeValue(maxValue - value)
								end,
	["Group Health"] =	function(frame, healthBar) 
							local maxValue = select(2, healthBar:GetMinMaxValues())
							if maxValue == 100 then 
								return ""
							end
							return FormatLargeValue(frame.aloftData.healthBarValue)
						end,
	["Group Health Deficit"] =	function(frame, healthBar) 
									local maxValue = select(2, healthBar:GetMinMaxValues())
									if maxValue == 100 then 
										return "" 
									end
									local value = frame.aloftData.healthBarValue
									if value == maxValue then 
										return "" 
									end
									return FormatLargeValue(maxValue - value)
								end,
	["Health"] =	function (frame, healthBar)
						local maxValue = select(2, healthBar:GetMinMaxValues())
						if maxValue ~= 100 then 
							return FormatLargeValue(frame.aloftData.healthBarValue)
						end

						if MobHealth3 then
							local health, maxHealth, valid = MobHealth3:GetHealth(frame.aloftData.healthBarValue, maxValue, frame.aloftData.name, frame.aloftData.level)
							if valid then
								return FormatLargeValue(health)
							end							
						elseif MobHealthDB then
							local mobHealthIndex = string.format("%s:%d", frame.aloftData.name, frame.aloftData.level)
							local mobHealthData = MobHealthDB[mobHealthIndex]
							if mobHealthData then
								local s, e, pts, pct = string.find(mobHealthData, "^(%d+)/(%d+)$")
								if( pts and pct ) then
									pct = tonumber(pct)
									if( pct ~= 0 ) then
										return FormatLargeValue((frame.aloftData.healthBarValue * tonumber(pts) / pct))
									end
								end							
							end
						end
						
						return string.format("%d%%", frame.aloftData.healthBarValue)
					end
}

-----------------------------------------------------------------------------

function Aloft:OnInitialize()
	self:SetText(L["Aloft"])
end

function Aloft:OnEnable()
	self:ScheduleRepeatingEvent("AloftUpdate", self.Update, 0.2, self)
	self:RegisterEvent("SharedMedia_SetGlobal", function(mediatype, override)
		if mediatype == "statusbar" then
			self:UpdateHealthBars()
			self:UpdateHighlights()
			self:UpdateOverlays()
			self:UpdateCastBars()
			self:UpdateCastBarOverlays()
		end
	end)
	
	self:CreateCastBarTextFrame()
	self:CreateCastBarTimeTextFrame()
end

function Aloft:OnDisable()
end

function Aloft:OnClick()
	if NAMEPLATES_ON then
		HideNameplates()
		NAMEPLATES_ON = false
	else
		ShowNameplates()
		NAMEPLATES_ON = true
	end
end

function Aloft:IsNameplateFrame(frame)
	if frame:GetName() then
		return false
	end

	local overlayFrame = frame:GetRegions()
	if not overlayFrame or overlayFrame:GetObjectType() ~= "Texture" or overlayFrame:GetTexture() ~= "Interface\\Tooltips\\Nameplate-Border" then
		return false
	end
	
	return true
end

function Aloft:ProcessChildFrames(...)
	for i = 1, select('#', ...) do
		local frame = select(i, ...)
		if self:IsNameplateFrame(frame) then 
			if not nameplateList[frame] then
				self:SetupNamePlate(frame)
				nameplateList[frame] = frame
			end
		end
	end
end

function Aloft:Update()
	local newNumberOfWorldChildren = WorldFrame:GetNumChildren()
	if newNumberOfWorldChildren ~= numberOfWorldChildren then
		numberOfWorldChildren = newNumberOfWorldChildren

		self:ProcessChildFrames(WorldFrame:GetChildren())
	end
end

if Prat_PlayerNames then
	Aloft.PratPlayerNamesUpdateMouseOver = Prat_PlayerNames.updateMouseOver
	Prat_PlayerNames.updateMouseOver = function()
		if UnitIsPlayer("mouseover") and UnitIsFriend("player", "mouseover") and not Prat_PlayerNames.Classes[UnitName("mouseover")] then
			Aloft.PratPlayerNamesUpdateMouseOver(Prat_PlayerNames)
			Aloft:UpdateHealthBarColourForUnitNamed(UnitName("mouseover"))
		end
	end

end

function Aloft:UpdateHealthBarColourForUnitNamed(name)
	for frame in pairs(nameplateList) do	
		if frame.aloftData.name == name then
			self:UpdateNameplate(frame)
		end
	end
end

-----------------------------------------------------------------------------

function Aloft:PlaceFrame(frame, healthBar, placement)
	if placement.visible then
		frame:ClearAllPoints()
		frame:SetPoint(placement.point, healthBar, placement.relativeToPoint, placement.offsetX, placement.offsetY)
		
		if placement.size then
			frame:SetWidth(placement.size)
			frame:SetHeight(placement.size)
		end
		frame:SetAlpha(placement.alpha or 1.0)
	else
		frame:SetAlpha(0)
	end
end

-----------------------------------------------------------------------------

function Aloft:UpdateHealthBarColor(frame, healthBar)
	if self.db.profile.healthBar.colorByClass then
		if frame.aloftData.type == "friendlyPlayer" then
			if not self.db.profile.healthBar.groupOnly and Prat_PlayerNames and Prat_PlayerNames.Classes[frame.aloftData.name] then		   
				local color = RAID_CLASS_COLORS[string.upper(Prat_PlayerNames.Classes[frame.aloftData.name])]
				if color then
					healthBar:SetStatusBarColor(color.r, color.g, color.b, self.db.profile.healthBar.alpha)
					return
				end
			end
			if roster then
				local unitid = roster:GetUnitIDFromName(frame.aloftData.name)
				if unitid then
					local color = RAID_CLASS_COLORS[select(2, UnitClass(unitid))]
					if color then
						healthBar:SetStatusBarColor(color.r, color.g, color.b, self.db.profile.healthBar.alpha)
						return
					end
				end
			end
		elseif frame.aloftData.type == "friendlyNPC" then
			if roster and roster:GetUnitIDFromName(frame.aloftData.name) then
				-- Must be a party pet
				local color = self.db.profile.healthBar.colors.partyPet
				healthBar:SetStatusBarColor(color[1], color[2], color[3], self.db.profile.healthBar.alpha)
				return
			end
		end
	end
	
	local color = self.db.profile.healthBar.colors[frame.aloftData.type]
	if color then
		healthBar:SetStatusBarColor(color[1], color[2], color[3], self.db.profile.healthBar.alpha)
	else
		local r, g, b = healthBar:GetStatusBarColor()
		healthBar:SetStatusBarColor(r, g, self.db.profile.healthBar.alpha)
	end
end

function Aloft:UpdateHealthText(frame, healthBar)
	local healthTextFrame = frame.aloftData.healthTextFrame
	
	if healthTextFrame then
		healthTextFrame:SetText(healthTextMethods[self.db.profile.healthText.mode](frame, healthBar))
	end
end

function Aloft:UpdateNameplate(frame)
	local healthBar = frame:GetChildren()

	self:UpdateHealthBarColor(frame, healthBar)
	self:UpdateHealthText(frame, healthBar)

	-- Check comment in OnHealthBarValueChanged()
	healthBar:SetValue(frame.aloftData.healthBarValue-0.001)
end

function Aloft:OnHealthBarValueChanged(frame, healthBar)
	if frame:IsShown() then
		local _, _, _, _, nameFrame = frame:GetRegions()
		
		local aloftData = frame.aloftData
		aloftData.healthBarValue = healthBar:GetValue()
		
		if aloftData.nameFrameText ~= nameFrame:GetText() then
			self:OnNameplateShow(frame)
		else
			self:UpdateNameplate(frame)
		end
	end
end

function Aloft:OnNameplateShow(frame)
	-- When a nameplate is first shown, grab some data about it
	
	local overlayFrame, _, _, highlightFrame, nameFrame, levelTextFrame, bossIconFrame = frame:GetRegions()
	local healthBar = frame:GetChildren()

	local aloftData = frame.aloftData

	if aloftData.nameFrameText ~= nameFrame:GetText() then
		aloftData.name = nameFrame:GetText()
	end
	aloftData.nameFrameText = aloftData.name
	aloftData.healthBarValue = healthBar:GetValue()
	aloftData.isBoss = bossIconFrame:IsShown()
	aloftData.level = tonumber(levelTextFrame:GetText()) or -1
	if not levelTextFrame:IsShown() then
		aloftData.level = -1
	end

	local r, g, b = healthBar:GetStatusBarColor()
	r, g, b = floor(r * 256), floor(g*256), floor(b*256)

	if r == 255 and g == 0 and b == 0 then
		if aloftData.isBoss then
			aloftData.type = "hostileBoss"
		else
			aloftData.type = "hostile"
		end
	elseif r == 0 and g == 0 and b == 255 then
		aloftData.type = "friendlyPlayer"
	elseif r == 0 and g == 255 and b == 0 then
		if aloftData.isBoss then
			aloftData.type = "friendlyBoss"
		else
			aloftData.type = "friendlyNPC"
		end
	elseif r == 255 and g == 255 and b == 0 then
		aloftData.type = "neutral"
	else
--		ChatFrame3:AddMessage("rgb = "..r.."."..g.."."..b)
	end

	if self.db.profile.nameplateText.overrideColors then
		local color = self.db.profile.nameplateText.colors[aloftData.type]
	
		if color then
			aloftData.nameFrameText = string.format("|c%02x%02x%02x%02x%s", color[4]*255, color[1]*255, color[2]*255, color[3]*255, aloftData.name)
			nameFrame:SetText(aloftData.nameFrameText)
		end
	
	end

	if not self.db.profile.levelText.visible then
		levelTextFrame:Hide()
	end

	self:PlaceFrame(bossIconFrame, healthBar, self.db.profile.bossIcon)
	highlightFrame:ClearAllPoints()
	highlightFrame:SetAllPoints(healthBar)

	-- The frame has an active region for detecting hits that is given roughly by this formula.
	frame:SetHeight(healthBar:GetHeight()*1.6)
	frame:SetWidth(healthBar:GetWidth()*1.05)
	
	self:UpdateNameplate(frame)
end

function Aloft:OnNameplateHide(frame)
	frame.aloftData.name = "[invalid]"
	frame.aloftData.type = "unknown"
end

function Aloft:GetTargetSpellName()
	local _, _, displayName = UnitCastingInfo("target")
	if displayName then
		return displayName
	end
	_, _, displayName = UnitChannelInfo("target")
	if displayName then
		return displayName
	end
	return ""
end

function Aloft:OnCastBarValueChanged(frame, castBar)
	local _, castBarOverlayFrame = frame:GetRegions()
	castBarOverlayFrame:SetVertexColor(unpack(self.db.profile.castBarOverlay.color))

	if castBarTimeTextFrame then
		local minValue, maxValue = castBar:GetMinMaxValues()
		local value = castBar:GetValue()

		if value > frame.aloftData.lastCastTimeValue then
			if value >= maxValue then
				castBarTimeTextFrame:SetText("")
				frame.aloftData.lastCastTimeValue = 0
			else
				castBarTimeTextFrame:SetText(string.format("%.2f", maxValue-value))
				frame.aloftData.lastCastTimeValue = value
			end
		else
			if value <= maxValue then
				castBarTimeTextFrame:SetText("")
				frame.aloftData.lastCastTimeValue = 0
			else
				castBarTimeTextFrame:SetText(string.format("%.2f", value- minValue))
				frame.aloftData.lastCastTimeValue = value
			end
		end
	end

	if castBarTextFrame and frame.aloftData.castBarStartTime ~= castBar:GetMinMaxValues() then
		castBarTextFrame:SetText(self:GetTargetSpellName())
	end
end

function Aloft:OnCastBarShow(frame, castBar)
	local healthBar = frame:GetChildren()
	local targetSpellName = self:GetTargetSpellName()
	local _, castBarOverlayFrame, spellIconFrame = frame:GetRegions()

--[[
	if self.db.profile.castBar.hideExpired then
		if not targetSpellName then
			ChatFrame3:AddMessage("Trying to hide expired...")
			castBar:Hide()	
			castBarOverlayFrame:Hide()	
			spellIconFrame:Hide()
			return
		end
	end
]]

	local verticalOffset = self.db.profile.castBar.offsets.vertical - healthBar:GetHeight()

	castBar:ClearAllPoints()
	castBar:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.castBar.offsets.left, verticalOffset)
	castBar:SetPoint("BOTTOMRIGHT", healthBar, "TOPRIGHT", self.db.profile.castBar.offsets.right, verticalOffset - self.db.profile.castBar.height)

	castBarOverlayFrame:ClearAllPoints()
	castBarOverlayFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.castBar.offsets.left, verticalOffset)
	castBarOverlayFrame:SetPoint("BOTTOMRIGHT", healthBar, "TOPRIGHT", self.db.profile.castBar.offsets.right, verticalOffset - self.db.profile.castBar.height)
	castBarOverlayFrame:SetVertexColor(unpack(self.db.profile.castBarOverlay.color))
	
	self:PlaceFrame(spellIconFrame, castBar, self.db.profile.spellIcon)

	if castBarTextFrame and self.db.profile.castBarText.visible then
		castBarTextFrame:SetText(targetSpellName)
		castBarTextFrame:ClearAllPoints()
		castBarTextFrame:SetPoint("TOPLEFT", castBar, "TOPLEFT", self.db.profile.castBarText.offsets.left, self.db.profile.castBarText.offsets.vertical)
		castBarTextFrame:SetPoint("BOTTOMRIGHT", castBar, "BOTTOMRIGHT", self.db.profile.castBarText.offsets.right, self.db.profile.castBarText.offsets.vertical)
		castBarTextFrame:SetAlpha(1.0)
		castBarTextFrame:Show()

		self:ScheduleRepeatingEvent("AloftUpdateCastBarTextAlpha", self.UpdateCastBarTextAlpha, 0, self, spellIconFrame)
		
		frame.aloftData.castBarStartTime = castBar:GetMinMaxValues()
	end
	
	if castBarTimeTextFrame and self.db.profile.castBarTimeText.visible then
		frame.aloftData.lastCastTimeValue = castBar:GetValue()
		castBarTimeTextFrame:ClearAllPoints()
		castBarTimeTextFrame:SetPoint("TOPLEFT", castBar, "TOPLEFT", self.db.profile.castBarTimeText.offsets.left, self.db.profile.castBarTimeText.offsets.vertical)
		castBarTimeTextFrame:SetPoint("BOTTOMRIGHT", castBar, "BOTTOMRIGHT", self.db.profile.castBarTimeText.offsets.right, self.db.profile.castBarTimeText.offsets.vertical)
		castBarTimeTextFrame:Show()
	end
end

function Aloft:UpdateCastBarTextAlpha(spellIconFrame)
	castBarTextFrame:SetAlpha(spellIconFrame:GetAlpha())
end

-----------------------------------------------------------------------------

local function OnNameplateShow()
	Aloft:OnNameplateShow(this)
end

local function OnNameplateHide()
	Aloft:OnNameplateHide(this)
end

local function OnHealthBarValueChanged()
	-- So how do we get the game to send us a message when
	-- it's decided to recycle frames internally?
	--
	-- What we do is flag this Status Bar with a value that is 
	-- not quite what it used to be. The game only ever sets
	-- integral values for the status bars.
	-- We set it to a little bit less.
	--
	-- If the game recycles, it triggers an event as the
	-- values have changed.
	--
	-- If the ValueChange event was caused by our trick
	-- then it won't have an integral value, so we skip out
	local value = this:GetValue()
	if value ~= floor(value) then 
		return 
	end
	
	Aloft:OnHealthBarValueChanged(this:GetParent(), this)
end

local function OnCastBarValueChanged()
	Aloft:OnCastBarValueChanged(this:GetParent(), this)
end

local function OnCastBarShow()
	Aloft:OnCastBarShow(this:GetParent(), this)
end

local function OnCastBarHide()
	Aloft:CancelScheduledEvent("AloftUpdateCastBarTextAlpha")
	if castBarTextFrame then
		castBarTextFrame:Hide()
	end
	if castBarTimeTextFrame then
		castBarTimeTextFrame:Hide()
	end
end

-----------------------------------------------------------------------------

function Aloft:SetupHealthTextFrame(frame, healthBar)
	if self.db.profile.healthText.mode == "Hidden" then
		return
	end
	
	local healthTextFrame = frame.aloftData.healthTextFrame
	
	if not healthTextFrame then
		healthTextFrame = frame:CreateFontString()
		frame.aloftData.healthTextFrame = healthTextFrame
	end
	
	healthTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.healthText.fontSize, outlineLookup[self.db.profile.healthText.outline])
	healthTextFrame:SetTextColor(unpack(self.db.profile.healthText.color))
	self:PlaceFrame(healthTextFrame, healthBar, self.db.profile.healthText)
end

function Aloft:SetupCastBarTextFrame(frame, castBar)
	if frame.aloftData.name == UnitName("target") then
		castBarTextFrame:ClearAllPoints()
		castBarTextFrame:SetPoint("TOPLEFT", castBar, "TOPLEFT", self.db.profile.castBarText.offsets.left, self.db.profile.castBarText.offsets.vertical)
		castBarTextFrame:SetPoint("BOTTOMRIGHT", castBar, "BOTTOMRIGHT", self.db.profile.castBarText.offsets.right, self.db.profile.castBarText.offsets.vertical)
		castBarTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.castBarText.fontSize, outlineLookup[self.db.profile.castBarText.outline])
		castBarTextFrame:SetJustifyH(self.db.profile.castBarText.alignment)
		castBarTextFrame:SetTextColor(unpack(self.db.profile.castBarText.color))
	end	
end

function Aloft:SetupCastBarTimeTextFrame(frame, castBar)
	if frame.aloftData.name == UnitName("target") then
		castBarTimeTextFrame:ClearAllPoints()
		castBarTimeTextFrame:SetPoint("TOPLEFT", castBar, "TOPLEFT", self.db.profile.castBarTimeText.offsets.left, self.db.profile.castBarTimeText.offsets.vertical)
		castBarTimeTextFrame:SetPoint("BOTTOMRIGHT", castBar, "BOTTOMRIGHT", self.db.profile.castBarTimeText.offsets.right, self.db.profile.castBarTimeText.offsets.vertical)
		castBarTimeTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.castBarTimeText.fontSize, outlineLookup[self.db.profile.castBarTimeText.outline])
		castBarTimeTextFrame:SetJustifyH(self.db.profile.castBarTimeText.alignment)
		castBarTimeTextFrame:SetTextColor(unpack(self.db.profile.castBarTimeText.color))
	end	
end

function Aloft:SetupNamePlate(frame)
	frame.aloftData = { type = "unknown", healthBarValue = 0, lastCastTimeValue = 0 }

	local overlayFrame, castBarOverlayFrame, spellIconFrame, highlightFrame, nameFrame, levelTextFrame, bossIconFrame, raidIconFrame = frame:GetRegions()
	local healthBar, castBar = frame:GetChildren()

	highlightFrame:ClearAllPoints()
	highlightFrame:SetAllPoints(healthBar)
	highlightFrame:SetVertexColor(unpack(self.db.profile.highlight.color))	
	highlightFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.highlight.texture))
	highlightFrame:SetBlendMode(self.db.profile.highlight.blendMode)
	
	nameFrame:ClearAllPoints()
	nameFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.nameplateText.offsets.left, self.db.profile.nameplateText.offsets.vertical)
	nameFrame:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", self.db.profile.nameplateText.offsets.right, self.db.profile.nameplateText.offsets.vertical)
	nameFrame:SetFont(NAMEPLATE_FONT, self.db.profile.nameplateText.fontSize, outlineLookup[self.db.profile.nameplateText.outline])
	nameFrame:SetJustifyH(self.db.profile.nameplateText.alignment)
	nameFrame:SetShadowOffset(0, 0)

	self:PlaceFrame(levelTextFrame, healthBar, self.db.profile.levelText)
	levelTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.levelText.fontSize, outlineLookup[self.db.profile.levelText.outline])
	levelTextFrame:SetShadowOffset(0, 0)

	self:PlaceFrame(raidIconFrame, healthBar, self.db.profile.raidIcon)
	self:PlaceFrame(bossIconFrame, healthBar, self.db.profile.bossIcon)

	overlayFrame:ClearAllPoints()
	overlayFrame:SetAllPoints(healthBar)
	overlayFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.overlay.texture))
	overlayFrame:SetVertexColor(unpack(self.db.profile.overlay.color))

	local verticalOffset = self.db.profile.castBar.offsets.vertical - healthBar:GetHeight()

	castBar:ClearAllPoints()
	castBar:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.castBar.offsets.left, verticalOffset)
	castBar:SetPoint("BOTTOMRIGHT", healthBar, "TOPRIGHT", self.db.profile.castBar.offsets.right, verticalOffset - self.db.profile.castBar.height)
	castBar:SetStatusBarTexture(SML:Fetch("statusbar", self.db.profile.castBar.texture))

	-- Trying to anchor the castbar overlay to castbar results in some very unexpected complete-game crashes (no error message, or anything)..
	castBarOverlayFrame:ClearAllPoints()
	castBarOverlayFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.castBar.offsets.left, verticalOffset)
	castBarOverlayFrame:SetPoint("BOTTOMRIGHT", healthBar, "TOPRIGHT", self.db.profile.castBar.offsets.right, verticalOffset - self.db.profile.castBar.height)
	castBarOverlayFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.castBarOverlay.texture))
	castBarOverlayFrame:SetVertexColor(unpack(self.db.profile.castBarOverlay.color))
	
	if not self.db.profile.overlay.visible then	
		overlayFrame:Hide()
	end
	
	if not self.db.profile.nameplateText.visible then
		nameFrame:Hide()
	end

	frame:SetScript("OnShow", OnNameplateShow)
	frame:SetScript("OnHide", OnNameplateHide)

	healthBar:SetScript("OnValueChanged", OnHealthBarValueChanged)
	healthBar:SetStatusBarTexture(SML:Fetch("statusbar", self.db.profile.healthBar.texture))
	
	castBar:SetScript("OnShow", OnCastBarShow)
	castBar:SetScript("OnHide", OnCastBarHide)
	castBar:SetScript("OnValueChanged", OnCastBarValueChanged)

	self:PlaceFrame(spellIconFrame, castBar, self.db.profile.spellIcon)

	self:SetupHealthTextFrame(frame, healthBar)
	self:SetupCastBarTextFrame(frame, castBar)
	
	self:OnNameplateShow(frame)
end

function Aloft:CreateCastBarTextFrame()
	if not self.db.profile.castBarText.visible then
		if castBarTextFrame then
			castBarTextFrame:Hide()
		end
		return
	end

	if not castBarTextFrame then
		castBarTextFrame = UIParent:CreateFontString()
	end
	
	castBarTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.castBarText.fontSize, outlineLookup[self.db.profile.castBarText.outline])
	castBarTextFrame:SetShadowOffset(1, -1)
	castBarTextFrame:SetJustifyH(self.db.profile.castBarText.alignment)
	castBarTextFrame:SetTextColor(unpack(self.db.profile.castBarText.color))
end

function Aloft:CreateCastBarTimeTextFrame()
	if not self.db.profile.castBarTimeText.visible then
		if castBarTimeTextFrame then
			castBarTimeTextFrame:Hide()
		end
		return
	end

	if not castBarTimeTextFrame then
		castBarTimeTextFrame = UIParent:CreateFontString()
	end
	
	castBarTimeTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.castBarTimeText.fontSize, outlineLookup[self.db.profile.castBarTimeText.outline])
	castBarTimeTextFrame:SetShadowOffset(1, -1)
	castBarTimeTextFrame:SetJustifyH(self.db.profile.castBarTimeText.alignment)
	castBarTimeTextFrame:SetTextColor(unpack(self.db.profile.castBarTimeText.color))
end

-----------------------------------------------------------------------------
-- Options methods. All of these need access to nameplateList
-----------------------------------------------------------------------------

function Aloft:UpdateNameplateText()
	for frame in pairs(nameplateList) do	
		local _, _, _, _, nameFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
	
		if self.db.profile.nameplateText.visible then
			nameFrame:Show()
		else
			nameFrame:Hide()
		end
	
		nameFrame:ClearAllPoints()
		nameFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.nameplateText.offsets.left, self.db.profile.nameplateText.offsets.vertical)
		nameFrame:SetPoint("BOTTOMRIGHT", healthBar, "BOTTOMRIGHT", self.db.profile.nameplateText.offsets.right, self.db.profile.nameplateText.offsets.vertical)
		nameFrame:SetFont(NAMEPLATE_FONT, self.db.profile.nameplateText.fontSize, outlineLookup[self.db.profile.nameplateText.outline])
		nameFrame:SetJustifyH(self.db.profile.nameplateText.alignment)
		
		if frame:IsShown() then
			self:OnNameplateShow(frame)
		end
	end
end

function Aloft:UpdateHealthBars()
	for frame in pairs(nameplateList) do	
		local overlayFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
		healthBar:SetStatusBarTexture(SML:Fetch("statusbar", self.db.profile.healthBar.texture))
		if frame.aloftData.name ~= "[Invalid]" then
			self:UpdateHealthBarColor(frame, healthBar)
		end
	end
end

function Aloft:UpdateHealthTexts()
	for frame in pairs(nameplateList) do
		local healthBar = frame:GetChildren()
		self:SetupHealthTextFrame(frame, healthBar)
		self:UpdateHealthText(frame, healthBar)
	end
end

function Aloft:UpdateLevelText()
	for frame in pairs(nameplateList) do
		local _, _, _, _, _, levelTextFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
		
		if self.db.profile.levelText.visible then
			levelTextFrame:Show()
		else
			levelTextFrame:Hide()
		end
		levelTextFrame:SetFont(NAMEPLATE_FONT, self.db.profile.levelText.fontSize, outlineLookup[self.db.profile.levelText.outline])
		self:PlaceFrame(levelTextFrame, healthBar, self.db.profile.levelText)
	end
end

function Aloft:UpdateRaidIcons()
	for frame in pairs(nameplateList) do	
		local _, _, _, _, _, _, _, raidIconFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()
	
		Aloft:PlaceFrame(raidIconFrame, healthBar, self.db.profile.raidIcon)
	end
end

function Aloft:UpdateBossIcons()
	for frame in pairs(nameplateList) do	
		local _, _, _, _, _, _, bossIconFrame = frame:GetRegions()
		local healthBar = frame:GetChildren()

		Aloft:PlaceFrame(bossIconFrame, healthBar, self.db.profile.bossIcon)
	end
end

function Aloft:UpdateSpellIcons()
	for frame in pairs(nameplateList) do	
		local _, _, spellIconFrame = frame:GetRegions()
		local _, castBar = frame:GetChildren()

		Aloft:PlaceFrame(spellIconFrame, castBar, self.db.profile.spellIcon)
	end
end

function Aloft:UpdateOverlays()
	for frame in pairs(nameplateList) do
		local overlayFrame = frame:GetRegions()
		if self.db.profile.overlay.visible then
			overlayFrame:Show()
		else
			overlayFrame:Hide()
		end
		overlayFrame:SetVertexColor(unpack(self.db.profile.overlay.color))
		overlayFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.overlay.texture))
	end
end

function Aloft:UpdateCastBars()
	for frame in pairs(nameplateList) do
		local _, castBarOverlayFrame = frame:GetRegions()
		local healthBar, castBar = frame:GetChildren()
		
		local verticalOffset = self.db.profile.castBar.offsets.vertical - healthBar:GetHeight()
		
		castBar:ClearAllPoints()
		castBar:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.castBar.offsets.left, verticalOffset)
		castBar:SetPoint("BOTTOMRIGHT", healthBar, "TOPRIGHT", self.db.profile.castBar.offsets.right, verticalOffset - self.db.profile.castBar.height)
		castBar:SetStatusBarTexture(SML:Fetch("statusbar", self.db.profile.castBar.texture))

		castBarOverlayFrame:ClearAllPoints()
		castBarOverlayFrame:SetPoint("TOPLEFT", healthBar, "TOPLEFT", self.db.profile.castBar.offsets.left, verticalOffset)
		castBarOverlayFrame:SetPoint("BOTTOMRIGHT", healthBar, "TOPRIGHT", self.db.profile.castBar.offsets.right, verticalOffset - self.db.profile.castBar.height)
	end
end

function Aloft:UpdateCastBarOverlays()
	for frame in pairs(nameplateList) do
		local _, castBarOverlayFrame = frame:GetRegions()
		castBarOverlayFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.castBarOverlay.texture))
	end
end

function Aloft:UpdateCastBarText()
	self:CreateCastBarTextFrame()
	for frame in pairs(nameplateList) do
		local _, castBar = frame:GetChildren()
		self:SetupCastBarTextFrame(frame, castBar)
	end
end

function Aloft:UpdateCastBarTimeText()
	self:CreateCastBarTimeTextFrame()
	for frame in pairs(nameplateList) do
		local _, castBar = frame:GetChildren()
		self:SetupCastBarTimeTextFrame(frame, castBar)
	end
end

function Aloft:UpdateHighlights()
	for frame in pairs(nameplateList) do
		local _, _, _, highlightFrame = frame:GetRegions()
		highlightFrame:SetVertexColor(unpack(self.db.profile.highlight.color))
		highlightFrame:SetTexture(SML:Fetch("statusbar", self.db.profile.highlight.texture))
		highlightFrame:SetBlendMode(self.db.profile.highlight.blendMode)
	end
end

-----------------------------------------------------------------------------
